# -*- coding: utf-8 -*-
from construct import *
from binascii import hexlify,unhexlify
from ZCL_FRAME import *

# For debug ...
class PrintContext(Construct):
	def _parse(self, stream, context, path):
		print(context)

#### Status #####################################################


Status = Enum(Int8ub,
	OK = 0x00,
	MalformedCommand             = 0x80,
	UnsupportedClusterCommand    = 0x81,
	UnsupportedGeneralCommand    = 0x82,
	UnsupportedAttribute         = 0x86,
	InvalidField                 = 0x87,
	InvalidValue                 = 0x88,
	InsufficientSpace            = 0x89,
	UnreportableAttribute        = 0x8c,
	BatchReportNoFreeSlot        = 0xc2,
	BatchReportInvalidTag        = 0xc3,
	BatchReportDuplicateTagLabel = 0xc4,
	BatchReportLabelOutOfRange   = 0xc5,
	default                      = Pass
)

#### DataType ###################################################
DataType = Enum(Int8ub,
	Boolean               = 0x10,
	General8              = 0x08,
	General16             = 0x09,
	General24             = 0x0A,
	General32             = 0x0B,
	General40             = 0x0C,
	General48             = 0x0d,
	Bitmap8               = 0x18,
	Bitmap16              = 0x19,
	UInt8                 = 0x20,
	UInt16                = 0x21,
	UInt32                = 0x23,
	Int8                  = 0x28,
	Int16                 = 0x29,
	Int32                 = 0x2B,
	UInt8Enum             = 0x30,
	SinglePrecision       = 0x39,
	CharString            = 0x42,
	ByteString            = 0x41,
	LongByteString        = 0x43,
	StructOrderedSequence = 0x4C,
	default               = Pass
)

#### Endpoint ###################################################
class EndPointAdapter(Adapter):
	# make specific EndPoint Encoding/Decoding
	
	def _encode(self, obj, context):
		assert ((obj >= 0) and (obj < 32))
		return ( ((obj & 0x07) << 5) | 0x10 | ((obj & 0x18) >> 2) | 0x01)
		
	def _decode(self, obj, context):
		#we work only on 7 bits 
		return ( ((obj & 0x70) >> 4) | ((obj & 0x03) << 3) ) 

EndPoint = EndPointAdapter(BitsInteger(7))

FrameCtrl = BitStruct(
	"EndPoint" / EndPoint,
	"Report" / Enum(Bit, Standard = 1 , Batch = 0),
 )
 
import base64
class BytesTostrBase64(Adapter):
	# make specific EndPoint Encoding/Decoding
	
	def _encode(self, obj, context):
		#TODO
		return 0
		
	def _decode(self, obj, context):
		obj_base64 = base64.b64encode(obj).decode()
		return obj_base64
 
BytesTostrBase64 = BytesTostrBase64(GreedyBytes)



import binascii
class BytesTostrHex(Adapter):
	# make specific EndPoint Encoding/Decoding
	
	def _encode(self, obj, context):
		#TODO
		return 0
		
	def _decode(self, obj, context):
		obj_hex = binascii.hexlify(obj).decode()
		return obj_hex
 
BytesTostrHex = BytesTostrHex(GreedyBytes)

#### Command ID #################################################
CommandID = Enum(Int8ub,
	ReadAttribute                      = 0x00,
	ReadAttributeResponse              = 0x01,
	WriteAttributeNoResponse           = 0x05,
	ConfigureReporting                 = 0x06,
	ConfigureReportingResponse         = 0x07,
	ReadReportingConfiguration         = 0x08,
	ReadReportingConfigurationResponse = 0x09,
	ReportAttributes                   = 0x0A,
	ReportAttributesAlarm              = 0x8A,
	ClusterSpecificCommand             = 0x50,
	default                            = Pass
)

#### Cluster ID #################################################
ClusterID = Enum(Int16ub,
	Basic             = 0x0000,
	OnOff             = 0x0006,
	SimpleMetering    = 0x0052,
	PowerQuality      = 0x8052,
	Occupancy         = 0x0406,
	Temperature       = 0x0402,
	Pressure          = 0x8008,
	RelativeHumidity  = 0x0405,
	AnalogInput       = 0x000C,
	BinaryInput       = 0x000F,
	Illuminance       = 0x0400,
	MultiStateOutput  = 0x0013,
	Configuration     = 0x0050,
	VolumeMeter       = 0x8002,
	SensO             = 0x8003,
	LoRaWAN           = 0x8004,
	MultiBinaryInput  = 0x8005,
	SerialInterface   = 0x8006, 
	SerialMasterSlave = 0x8007,
	TIC_ICE           = 0x0053,
	TIC_CBE           = 0x0054,
	TIC_CJE           = 0x0055,
	TIC_STD           = 0x0056,
	TIC_PMEPMI        = 0x0057,
	default           = Pass
)


#### Attributes ID/Type (per cluster) ##############################
AttributeID = Switch(
	this._.ClusterID, {
		"Basic": Enum (Int16ub,
			FirmwareVersion     = 0x0002, 
			KernelVersion       = 0x0003,
			Manufacturer        = 0x0004,
			ModelIdentifier     = 0x0005,
			DateCode            = 0x0006,
			LocationDescription = 0x0010,
			ApplicationName     = 0x8001,
			default = Pass
		),
		"OnOff": Enum (Int16ub,
			State	= 0x0000,
			default = Pass
		),
		"SimpleMetering": Enum (Int16ub,
			CurrentMetering     = 0x0000,
			CurrentCalibration  = 0x8000,
			default = Pass
		),
		"PowerQuality": Enum (Int16ub,
			CurrentValues        = 0x0000,
			SagCycleThreshold    = 0x0001,
			SagVoltageThreshold  = 0x0002,
			OverVoltageThreshold = 0x0003,
			default = Pass
		),
		"Occupancy": Enum (Int16ub,
			Occupancy                    = 0x0000,
			OccupancyType                = 0x0001,
			PIROccupiedToUnoccupiedDelay = 0x0010,
			PIRUnoccupiedToOccupiedDelay = 0x0011,
			default = Pass
			
		),
		"Temperature": Enum (Int16ub,
			MeasuredValue    = 0x0000,
			MinMeasuredValue = 0x0001,
			MaxMeasuredValue = 0x0002,
			default = Pass
		),
		"Pressure": Enum (Int16ub,
			MeasuredValue    = 0x0000,
			MinMeasuredValue = 0x0001,
			MaxMeasuredValue = 0x0002,
			MeasurementPeriodAttribute =0x0003,
			SamplePerMeasurementAttribute =0x0004,
			SamplePerConfirmationMeasurementAttribute =0x0005,
			SamplePeriodeAttribute =0x0006,
			MeanMeasuredValueSinceLastReportAttribute =0x0100,
			MinimalMeasuredValueSinceLastReportAttribute =0x0101,
			MaximalMeasuredValueSinceLastReportAttribute =0x0102,
		    default = Pass
		),
		"RelativeHumidity": Enum (Int16ub,
			MeasuredValue    = 0x0000,
			MinMeasuredValue = 0x0001,
			MaxMeasuredValue = 0x0002,
			default = Pass
		),
		"AnalogInput": Enum (Int16ub,
			PresentValue    = 0x0055,
			ApplicationType = 0x0100,
			default = Pass
		),
		"BinaryInput": Enum (Int16ub,
			PresentValue    = 0x0055,
			Polarity        = 0x0054,
			ApplicationType = 0x0100,
			EdgeSelection   = 0x0400,
			DebouncePeriod  = 0x0401,
			Count           = 0x0402,
			default = Pass
		),
		"Illuminance": Enum (Int16ub,
		default = Pass),
		"MultiStateOutput": Enum (Int16ub,
			PresentValue    = 0x0055,
			NumberOfStates  = 0x004A,
			ApplicationType = 0x0100,
			default = Pass
		),
		"Configuration": Enum (Int16ub,
			Descriptor          = 0x0004,
			ConfigurationMode   = 0x0005,
			NodePowerDescriptor = 0x0006,
			default = Pass
		),
		"VolumeMeter": Enum (Int16ub,
			Volume            = 0x0000,
			VolumeDisplayMode = 0x0001,
			MinFlow           = 0x0002,
			MaxFlow           = 0x0003,
			FlowDisplayMode   = 0x0004,
			default = Pass
		),
		"SensO": Enum (Int16ub,
			Status                 = 0x0000,
			CountDownThresholds    = 0x0001,
			InstallationRotation   = 0x0002,
			VolumeRotation         = 0x0003,
			TemperatureMeterFreeze = 0x0004,
			TemperatureMinTxOff    = 0x0005,
			ParametersLeakFlow     = 0x0006,
			default = Pass
		),
		"LoRaWAN": Enum (Int16ub,
			MessageType              = 0x0000,
			NumberOfRetryIfConfirmed = 0x0001,
			ReAssociationParameters  = 0x0002,
			DataRateParameters       = 0x0003,
			ABPDevAddr               = 0x0004,
			OTAAppEUI                = 0x0005,
			default = Pass
		),
		"MultiBinaryInput": Enum (Int16ub,
			PresentValues = 0x0000,
			default = Pass
		),
		"SerialInterface": Enum (Int16ub,
			Speed    = 0x0000,
			DataBits = 0x0001,
			Parity   = 0x0002,
			StopBits = 0x0003,
			default = Pass
		), 
		"SerialMasterSlave": Enum (Int16ub,
			Request         = 0x0000,
			Answer          = 0x0001,
			ApplicationType = 0x0002,
			default = Pass
		),
		"TIC_ICE": Enum (Int16ub,
			Attribute_0        = 0x0000,
			Attribute_1        = 0x0001,
			Attribute_2        = 0x0002,
			Attribute_virtual_0        = 0x0100,
			Attribute_virtual_1        = 0x0101,
			Attribute_virtual_2        = 0x0102,
			default = Pass
		),
		"TIC_CBE": Enum (Int16ub,
			Attribute_0        = 0x0000,
			Attribute_virtual_0        = 0x0100,
			Attribute_virtual_1        = 0x0200,
			Attribute_virtual_2        = 0x0300,
			Attribute_virtual_3        = 0x0400,
			Attribute_virtual_4        = 0x0500,
			default = Pass
		),
		"TIC_CJE": Enum (Int16ub,
			Attribute_0        = 0x0000,
			Attribute_virtual_0        = 0x0100,
			Attribute_virtual_1        = 0x0200,
			Attribute_virtual_2        = 0x0300,
			Attribute_virtual_3        = 0x0400,
			Attribute_virtual_4        = 0x0500,
			default = Pass
		),
		"TIC_STD": Enum (Int16ub,
			Attribute_0        = 0x0000,
			Attribute_virtual_0        = 0x0100,
			Attribute_virtual_1        = 0x0200,
			Attribute_virtual_2        = 0x0300,
			Attribute_virtual_3        = 0x0400,
			Attribute_virtual_4        = 0x0500,
			default = Pass
		),
		"TIC_PMEPMI": Enum (Int16ub,
			Attribute_0        = 0x0000,
			Attribute_virtual_0        = 0x0100,
			Attribute_virtual_1        = 0x0200,
			Attribute_virtual_2        = 0x0300,
			Attribute_virtual_3        = 0x0400,
			Attribute_virtual_4        = 0x0500,
			default = Pass
		)
	},default = "Bytes" / GreedyBytes 
) 


#### Data (According to Attribute Type) ##############################
Data = Switch(
	this._.AttributeType, {
		"Boolean"              : Flag,
		"General8"             : Byte,
		"General16"            : Bytes(2),
		"General24"            : Bytes(3),
		"General32"            : Bytes(4),
		"General40"            : Bytes(4),
		"General48"            : 
			Switch(this._.AttributeID, {
				"FirmwareVersion" : Struct(
					"Major" / Int8ub,
			        "Minor" / Int8ub,
			        "Revision" / Int8ub,
			        "Build" / Int24ub,
			    )
			},	default =   "Bytes" / BytesTostrHex
				
			),
		"Bitmap8"              : BitStruct("b7" / BitsInteger(1),"b6" / BitsInteger(1),"b5" / BitsInteger(1),"b4" / BitsInteger(1),"b3" / BitsInteger(1),"b2" / BitsInteger(1),"b1" / BitsInteger(1),"b0" / BitsInteger(1)),
		"Bitmap16"             : BitStruct("b15" / BitsInteger(1),"b14" / BitsInteger(1),"b13" / BitsInteger(1),"b12" / BitsInteger(1),"b11" / BitsInteger(1),"b10" / BitsInteger(1),"b9" / BitsInteger(1),"b8" / BitsInteger(1),"b7" / BitsInteger(1),"b6" / BitsInteger(1),"b5" / BitsInteger(1),"b4" / BitsInteger(1),"b3" / BitsInteger(1),"b2" / BitsInteger(1),"b1" / BitsInteger(1),"b0" / BitsInteger(1)),
		"UInt8"                : Int8ub,
		"UInt16"               : Int16ub,
		"UInt32"               : Int32ub,
		"Int8"                 : Int8sb,
		"Int16"                : Int16sb,
		"Int32"                : Int32sb,
		"UInt8Enum"            : Int8ub,
		"SinglePrecision"               : Float32b,
		"CharString"           : 
			Struct(
			   "Count"  / Int8ub, 
			   "String" / String(this.Count)
			 ),
		"ByteString"           : 
			Switch(this._.AttributeID, {
				"CurrentMetering" : Struct(
					Const(Int8ub,12), 
			        "ActiveEnergy" / Int24sb,
			        "ReactiveEnergy" / Int24sb,
			        "NbMinutes" / Int16ub,
			        "ActivePower" / Int16sb,
			        "ReactivePower" / Int16sb,
			    ),
				"NodePowerDescriptor" : Struct(
					Int8ub,
			        "CurrentPowerMode" / Enum(Int8ub, ONWhenIdle = 0 , PeriodicallyON = 1, ONOnUserEvent = 2),
			        "AvailablePowerSourceBitField" / Int8ub,
					Embedded ( 	If ( ((this.AvailablePowerSourceBitField & 0x01 == 0x01)), Struct("ConstantVoltage" / Int16ub))),
					Embedded ( 	If ( ((this.AvailablePowerSourceBitField & 0x02 == 0x02)), Struct("RechargeableBatteryVoltage" / Int16ub))),
					Embedded ( 	If ( ((this.AvailablePowerSourceBitField & 0x04 == 0x04)), Struct("DisposableBatteryVoltage" / Int16ub))),
					Embedded ( 	If ( ((this.AvailablePowerSourceBitField & 0x08 == 0x08)), Struct("SolarHarvestingVoltage" / Int16ub))),
					Embedded ( 	If ( ((this.AvailablePowerSourceBitField & 0x10 == 0x10)), Struct("TicHarvestingVoltage" / Int16ub))),
			        "CurrentPowerSource" / Enum(Int8ub, Constant = 1 , RechargeableBattery = 2, DisposableBattery = 4, SolarHarvesting = 8, TicHarvesting = 16)
			    )
			},	default = Struct(
					"Count" / Int8ub, 
			        "Bytes" / BytesTostrHex
				)
			),
		"LongByteString"       : Struct(
			                       "Count" / Int16ub, 
			                       "Bytes" / BytesTostrHex
			                     ),
		"StructOrderedSequence": BytesTostrHex
	},default = "Bytes" / BytesTostrHex
)


	


Direction = Byte
MinReport = BitStruct(
	"Unit" / Enum(Bit, Minutes = 1 , Seconds = 0),
	"Value" / BitsInteger(15) 
 )
MaxReport = BitStruct(
	"Unit" / Enum(Bit, Minutes = 1 , Seconds = 0),
	"Value" / BitsInteger(15) 
 )



